from math import cos
from numpy import (
   linspace)

from file1 import (
    foo)

def bar(x):
    return x

class MyOtherClass:
    D = 1

    def three(self):
        return 3

    def four(self):
        return 4
