# %% first Code Cell's name
a = 10
"Hello"

# In[23]
li = [1, 2, 3]

# <codecell> Third Code Cell\'s name
import numpy as np
arr = np.array(li)

#%% Fourth Code/ Cell's % name\
s = """Z:\\escape\\test\\string\n"""
a
